/*
	HERUN.C

	contains Run...() routines

	for the Hugo Engine

	Copyright (c) 1995-1998 by Kent Tessman
*/


/* The only difference between this and the 'standard' herun.c
   is in the replacement of the call to hugo_iskeywaiting,
   which I could no satisfactorily inplement using the EMX
   library */

#include "heheader.h"

/* Function prototypes: */
int GetVal(void);               /* from heexpr.c */

void hugo_stopmusic(void);
void hugo_stopsample(void);


/* Used by RunWindow() for setting current window dimensions: */
int last_window_top, last_window_bottom, last_window_left, last_window_right;
int lowest_windowbottom = 0,			/* in text lines */
	physical_lowest_windowbottom;		/* in pixels or text lines */


/* RUNDO */

void RunDo(void)
{
	long skip, enterptr;

	enterptr = ++codeptr;
	skip = PeekWord(codeptr);       /* remember the skip distance */
	codeptr+=2;

	SetStackFrame(stack_depth, DOWHILE_BLOCK, skip+enterptr, codeptr);

#if defined (DEBUGGER)
	dbnest++;
#endif
}


/* RUNEVENTS */

void RunEvents(void)
{
	int i, tempundo, flag;
	int eventin, tempself;
	int templocals[MAXLOCALS];
	int temp_stack_depth;
	long tempptr, eventaddr;

	tempundo = undorecord;
	undorecord = true;

	tempptr = codeptr;
	tempself = var[self];

	temp_stack_depth = stack_depth;

	for (i=0; i<MAXLOCALS; i++)
		templocals[i] = var[MAXGLOBALS+i];

	for (i=0; i<events; i++)
	{
		defseg = eventtable;

		eventin = PeekWord(2 + i * 4);
		eventaddr = (long)PeekWord(2 + i * 4 + 2)*address_scale;
		var[self] = eventin;

		domain = 0, flag = 0;

		z = 1;

		if (eventin==0 or GrandParent(eventin)==GrandParent(var[player]))
			flag = 1;

		else if (Available(eventin))
			flag = 1;

		if (flag)
		{
			PassLocals(0);

			SetStackFrame(stack_depth, RUNROUTINE_BLOCK, 0, 0);
#if defined (DEBUGGER)
			DebugRunRoutine(eventaddr);
#else
			RunRoutine(eventaddr);
#endif
			stack_depth = temp_stack_depth;

			retflag = 0;
			if (var[endflag]) break;
		}
	}

	for (i=0; i<MAXLOCALS; i++)
		var[MAXGLOBALS+i] = templocals[i];

	codeptr = tempptr;
	var[self] = tempself;
	undorecord = (char)tempundo;
}


/* RUNGAME */

extern char reparse_everything;		/* from ParseError() in heparse.c */

void RunGame(void)
{
	char jw = 0;                    /* just wrote undo info */
	char wasxverb = 0, newinput;
	int i, flag, mc, lastspeaking = 0, startlocation;
	
	/* Set up initial screen position */
	hugo_settextpos(1, physical_windowheight/lineheight);
	full = 0;

	/* Load globals */
	defseg = arraytable;
	for (i=0; i<MAXGLOBALS; i++)
		var[i] = PeekWord(i*2);

	/* Reset the speaking-to variable */
	speaking = 0;

	if (game_version < 22)
		{passlocal[0] = objects;
		passlocal[1] = physical_windowwidth/FIXEDCHARWIDTH;}

#if defined (DEBUGGER)
RestartDebugger:

	/* If no gamefile is loaded, jump immediately to the debugger
	   interrupt function.
	*/
	if (game==NULL) Debugger();
#endif


	stack_depth = RESET_STACK_DEPTH;

	InitGame();

	undoptr = 0;
	undoturn = 0;
	undoinvalid = 1;
	undorecord = 0;

Start:
	stack_depth = 0;

	strcpy(errbuf, "");
	strcpy(oops, "");

	do
	{
		if (xverb==0)
		{
			undorecord = true;

			SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

			PassLocals(0);

#if defined (DEBUGGER)
			currentroutine = mainaddr;
			window[VIEW_CALLS].count = 0;

			/* If there is no Main routine */
			if (currentroutine==0) goto NormalTermination;

			DebugRunRoutine((long)mainaddr*address_scale);
#else
			RunRoutine((long)mainaddr*address_scale);
#endif
			undorecord = false;
			if (retflag)
				break;
		}

		if (!undoinvalid and !wasxverb)
		{
			undorecord = true;
			SaveUndo(0, undoturn, 0, 0, 0);
			undorecord = false;
			undoturn = 0;
		}
		else if (undoinvalid)
		{
			undoptr = 0;
			undoinvalid = 0;
			undoturn = 0;
			undorecord = true;
			SaveUndo(0, 0, 0, 0, 0);
			SaveUndo(0, 0, 0, 0, 0);
			undorecord = false;
		}

		xverb = true;

		jw = 1;
		if (var[endflag])
			break;
		jw = 0;

		z = 0;

	/* Endless game loop begins here */

		do
		{
			PassLocals(0);

			/* If there's nothing waiting to be finished in the
			   input buffer:
			*/
			if (!remaining)
			{
				newinput = true;
				speaking = 0;
				do
				{
FreshInput:
					if (z != true)
					{
#if defined (DEBUGGER)
						AddStringtoCodeWindow("[Waiting for input]");
						runaway_counter = 0;
						buffered_code_lines = FORCE_REDRAW;
						debugger_has_stepped_back = false;
						window[VIEW_LOCALS].changed = true;
#endif
						if (!playback)
						{
							GetCommand();
						}
						else
						{
							if (!fgets(buffer, MAXBUFFER, playback))
							{
								if (fclose(playback))
									FatalError(READ_E);
								playback = 0;
								GetCommand();
							}
							else
							{
								buffer[strlen(buffer)-1] = '\0';
								sprintf(line, "\n%s%s", GetWord(var[prompt]), buffer);
								if (script)
									if (fprintf(script, "\n") < 0)
										FatalError(WRITE_E);
#if defined (SCROLLBACK_DEFINED)
								hugo_sendtoscrollback("\n");
#endif
								AP(line);
							}
						}

#if defined (DEBUGGER)
						if (debugger_collapsing)
							goto NormalTermination;
#endif

						SeparateWords();

						if (record)
						{
							for (i=1; i<=words; i++)
							{
								if (!strcmp(word[i], "."))
								{
									if (fprintf(record, "\n")<0)
										FatalError(WRITE_E);
									if (i==words) goto RecordedNewline;
								}
								else if (fprintf(record, "%s ", word[i])<0)
										FatalError(WRITE_E);

							}
							if (fprintf(record, "\n")<0) FatalError(WRITE_E);
RecordedNewline:;
						}
					}
					else z = 0;

					if (!strcmp(buffer, "") or buffer[0]=='.')
					{
						strcpy(parseerr, "");

						/* "What?" */
						ParseError(0, 0);
						goto FreshInput;
					}
				}

				/* Loop until valid input */
				while (Parse()==false and strcmp(buffer, ""));
			}


			/* Else if there's something left in the input buffer */
			else
			{
				newinput = false;

				/* Erase the just-parsed command, and check to
				   to see if what's left is just blanks
				*/
				while (words > remaining)
					KillWord(1);
				flag = false;
				for (i=1; i<=words; i++)
					if (wd[i]!=0) flag = true;
				if (!flag) goto FreshInput;

				if (words) AP("");

				if (Parse()==false)
					{mc = false;
					goto Skipmc;}
			}


			/* Run the user Parse routine if one exists */
			if (parseaddr)
			{
				SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

				ret = 0;
				PassLocals(0);
#if defined (DEBUGGER)
				DebugRunRoutine((long)parseaddr*address_scale);
#else
				RunRoutine((long)parseaddr*address_scale);
#endif
				retflag = 0;

				/* Returning non-zero return calls the
				   engine's Parse routine again.
				*/
				if (ret) Parse();
			}
			reparse_everything = false;
			do
			{
				mc = MatchCommand();
				if (mc==false)
					remaining = 0;
			} while (reparse_everything and !mc);
Skipmc:;
		}
		while (!mc);

		if (!xverb) undorecord = true;

		wasxverb = xverb;

		/* If there's an unknown string to be put in parse$ */
		if (parsestr[0]!='\0') strcpy(parseerr, parsestr);
		else strcpy(parseerr, "");

		var[actor] = var[player];

		if (!newinput and lastspeaking) speaking = lastspeaking;

		if (var[verbroutine]!=0 or (speaking))
		{
			/* If command addresses an object/char. */
			if (speaking)
			{
				lastspeaking = speaking;

				/* If user Speakto routine exists */
				if (speaktoaddr)
				{
					var[object] = objlist[0];
					ret = 0;

					SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

					passlocal[0] = speaking;
					PassLocals(1);
#if defined (DEBUGGER)
					DebugRunRoutine((long)speaktoaddr*address_scale);
#else
					RunRoutine((long)speaktoaddr*address_scale);
#endif
					if (ret==0)
					{
						remaining = 0;
						xverb = true;
					}
					retflag = 0;
				}

				else
				{
					/* "...start with a verb..." */
					ParseError(2, 0);
					xverb = true;
				}
			}

			/* Regular old vanilla command: */
			else
			{
				speaking = 0;
				lastspeaking = 0;

	/* As of v2.5, the Perform junction routine takes care of calling the
	   before routines, verbroutine, etc.
	*/
				if (game_version>=25 and performaddr!=0)
				{
					i = 0;
NextPerform:
					if (objcount) var[object] = objlist[i];

					/* Have to do this before passing locals, in case
					   Name() ends up calling a routine (which would
					   trash passlocal[])
					*/
					if (parseerr[0]=='\0')
						strcpy(parseerr, Name(objlist[i]));

					/* Set up arguments for Perform */
					passlocal[0] = var[verbroutine];
					passlocal[1] = var[object];
					passlocal[2] = var[xobject];

					/* 'queue' argument, >1 if objcount > 1 */
					passlocal[3] = (objcount>1)?(i+1):0;
					/* -1 if object is a digit */
					if (object_is_number) passlocal[3] = (short)-1;

					obj_match_state = -1;
					startlocation = var[location];
					ret = 0;

					SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

					PassLocals(4);
#if defined (DEBUGGER)
					DebugRunRoutine((long)performaddr*address_scale);
#else
					RunRoutine((long)performaddr*address_scale);
#endif
					if (ret==0)
					{
						remaining = 0;
						xverb = true;
					}
					retflag = 0;

					/* Break if endflag is set or if the location has
					   changed from the first call to Perform
					*/
					if (var[endflag] or startlocation!=var[location])
						goto EndofCommand;

					if (objcount>1 and ++i<objcount)
						goto NextPerform;
			    	}

	/* v2.4 or earlier had to call the verb loop via the engine */

				/* One or more objects specified */
				else if (objcount > 0)  /* "if (objcount > 0" for pre-v2.5 */
				{
				  obj_match_state = 1;
				  startlocation = var[location];
				  for (i=0; i<objcount; i++)
				  {
				    if (parseerr[0]=='\0')
				    	strcpy(parseerr, Name(objlist[i]));

				    if (ValidObj(objlist[i]) and
					((objcount>1 and objlist[i]!=var[xobject]) or objcount==1))
				    {
				      var[object] = objlist[i];
				      if (GetProp(var[player], before, 1, 0)==0)
				      {
					if (GetProp(var[location], before, 1, 0)==0)
					{
					  if (GetProp(var[xobject], before, 1, 0)==0)
					  {
				  	    /* If multiple objects are specified, print
					       "name:  " for each:
					    */
					    if (objcount > 1)
					    {
					      sprintf(line, "%s:  \\;", Name(var[object]));
					      AP(line);
					    }

					    obj_match_state = 0;

					    if ((object_is_number) or GetProp(var[object], before, 1, 0)==0)
					    {
					      obj_match_state = -1;
					      ret = 0;

					      SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

					      PassLocals(0);
#if defined (DEBUGGER)
					      DebugRunRoutine((long)var[verbroutine]*address_scale);
#else
					      RunRoutine((long)var[verbroutine]*address_scale);
#endif
					      if (ret==0)
					      {
						remaining = 0;
						xverb = true;
					      }
					      retflag = 0;

					      GetProp(var[player], after, 1, 0);
					      GetProp(var[location], after, 1, 0);
					    }
					  }
				        }
				      }
				    }
				    if (var[endflag] or var[location]!=startlocation)
					break;
				  }
				}

				/* No object(s) specified */
				else
				{
				  if (GetProp(var[player], before, 1, 0)==0)
				  {
				    if (GetProp(var[location], before, 1, 0)==0)
				    {
				      ret = 0;

				      SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

				      PassLocals(0);
#if defined (DEBUGGER)
				      DebugRunRoutine((long)var[verbroutine]*address_scale);
#else
				      RunRoutine((long)var[verbroutine]*address_scale);
#endif
				      if (ret==0)
				      {
					remaining = 0;
					xverb = true;
				      }
				      retflag = 0;

				      GetProp(var[player], after, 1, 0);
				      GetProp(var[location], after, 1, 0);
				    }
		        	  }
			        }

	/* (end of pre-v2.5 verbroutine-calling) */

			}
		}
EndofCommand:
		if (var[endflag])
			break;

		undorecord = false;
	}
	while (true);   /* endless loop back to start */

	undorecord = false;

	if (var[endflag]==-1)
#if defined (DEBUGGER)
		goto NormalTermination;
#else
		return;
#endif

	if (!jw)
	{
		undorecord = true;
		SaveUndo(0, undoturn, 0, 0, 0);
		undorecord = false;
		undoturn = 0;
	}

	if (playback)
	{
		if (fclose(playback)) FatalError(READ_E);
		playback = 0;
	}

	Flushpbuffer();


	/* Run the user Endgame routine if one exists */

#if defined (DEBUGGER)
	if (endgameaddr and !debugger_collapsing)
#else
	if (endgameaddr)
#endif
	{
		passlocal[0] = var[endflag];

		SetStackFrame(RESET_STACK_DEPTH, RUNROUTINE_BLOCK, 0, 0);

		ret = 0;
		var[endflag] = 0;
		PassLocals(0);
#if defined (DEBUGGER)
		DebugRunRoutine((long)endgameaddr*address_scale);
#else
		RunRoutine((long)endgameaddr*address_scale);
#endif
		retflag = false;
	}
	else
		ret = 0;

	/* Stop all audio after running EndGame */
	hugo_stopmusic();
	hugo_stopsample();

	xverb = true;
	wasxverb = true;
	if (ret) goto Start;

#if defined (DEBUGGER)

NormalTermination:

	/* Normal program termination doesn't exit the debugger. */

	debugger_interrupt = true;
	debugger_run = false;
	var[endflag] = false;

	xverb = false;

	SwitchtoDebugger();
	UpdateDebugScreen();

	if (debugger_collapsing!=2)
	{
		DebugMessageBox("Program Exiting", "Normal program termination");

	}

	debugger_collapsing = false;

	if ((game!=NULL) and !RunRestart())
		DebugMessageBox("Restart Error", "Unable to restart");

	SwitchtoGame();

	history_count = 0;
	window[VIEW_CALLS].count = 0;

	for (i=0; i<(int)window[CODE_WINDOW].count; i++)
		free(codeline[i]);
	window[CODE_WINDOW].count = 0;

	/* Force Code window redraw */
	buffered_code_lines = FORCE_REDRAW;

	goto RestartDebugger;
#endif
}


/* RUNIF */

void RunIf(char override)
{
	char t, tempinexpr;
	long enterptr, skip;

	switch (t = mem[codeptr])
	{
		case CASE_T:
		case IF_T:
		case ELSEIF_T:
		case WHILE_T:
		case FOR_T:
		{
			codeptr++;
			enterptr = codeptr;

			/* Remember the skip distance */
			skip = PeekWord(codeptr);
			codeptr += 2;

			/* Check if we've already done an elseif */
			if (override and t==ELSEIF_T)
			{
				codeptr = skip+enterptr;
				return;
			}

			/* Read the expression */
			tempinexpr = inexpr;
			inexpr = 1;
			SetupExpr();
			inexpr = tempinexpr;

			/* If the expression is false, skip the
			   conditional block
			*/
			if (EvalExpr(0)==0)
			{
				codeptr = skip+enterptr;
				return;
			}

			/* Protect the stack if jumping backward */
			if (mem[codeptr]==JUMP_T)
				if ((long)(PeekWord(codeptr+1)*address_scale) < codeptr)
					if (--stack_depth < 0) stack_depth = 0;

			/* Continue on into the conditional block if
			   the expression evaluated to non-zero
			*/
PasstoBlock:
			if (t==WHILE_T or t==FOR_T)
				SetStackFrame(stack_depth, CONDITIONAL_BLOCK, skip+enterptr, 0);

			else	/* no 'break' parameter */
				SetStackFrame(stack_depth, CONDITIONAL_BLOCK, 0, 0);
#if defined (DEBUGGER)
			dbnest++;
#endif
			return;
		}
		case ELSE_T:
		{
			skip = PeekWord(++codeptr);
			enterptr = codeptr;
			codeptr += 2;

			if (override)
			{
				codeptr = skip+enterptr;
				return;
			}

			if (mem[codeptr]==JUMP_T)
				if ((long)(PeekWord(codeptr+1)*address_scale) < codeptr)
					if (--stack_depth < 0) stack_depth = 0;

			goto PasstoBlock;
		}
	}
}


/* RUNINPUT */

void RunInput(void)
{
	int i;

	strcpy(parseerr, "");

	Flushpbuffer();

	hugo_getline("");

#if defined (DEBUGGER)
	if (debugger_collapsing) return;
#endif

	strcpy(buffer, Rtrim(strlwr(buffer)));

	SeparateWords();

	for (i=1; i<=words; i++)
	{
		wd[i] = FindWord(word[i]);

		/* If a word isn't in the dictionary */
		if (wd[i]==UNKNOWN_WORD)
		{
			wd[i] = 0;
			strcpy(parseerr, word[i]);
			if (parseerr[0]=='\"')
			{
				strcpy(parseerr, Right(parseerr, strlen(parseerr)-1));
				if (parseerr[strlen(parseerr)-1]=='\"')
					parseerr[strlen(parseerr)-1] = '\0';
			}
		}
	}
	remaining = 0;
}


/* RUNMOVE */

/* (All the debugger range-checking is important because invalid memory
   writes due to invalid object location calculations are a good way to
   crash the system.)
*/

void RunMove(void)
{
	int obj, p;
#if defined (DEBUGGER)
	char out_of_range = 0;
#endif

	switch (mem[codeptr])
	{
		case MOVE_T:
		{
			codeptr++;
			obj = GetValue();

#if defined (DEBUGGER)
			if (!CheckinRange(obj, objects, "object"))
				out_of_range = true;
			else
#endif
			SaveUndo(MOVE_T, obj, Parent(obj), 0, 0);

			codeptr++;              /* skip "to" */
			p = GetValue();

#if defined (DEBUGGER)
			if (!CheckinRange(p, objects, "object"))
				out_of_range = true;

			if (!out_of_range)
#endif
			MoveObj(obj, p);
			break;
		}

		case REMOVE_T:
		{
			codeptr++;
			obj = GetValue();

#if defined (DEBUGGER)
			if (!CheckinRange(obj, objects, "object"))
				out_of_range = true;

			else
#endif
			SaveUndo(MOVE_T, obj, Parent(obj), 0, 0);

#if defined (DEBUGGER)
			if (!out_of_range)
#endif
			MoveObj(obj, 0);        /* move to parent 0 */
			break;
		}
	}

	if (game_version>=23) codeptr++; /* eol */
}


/* RUNPRINT */

void RunPrint(void)
{
	char number = 0, hexnumber = 0;
	int a;
	int i, l;

/*	capital = 0; */

	codeptr++;

	while (mem[codeptr] != EOL_T)
	{
		strcpy(line, "");

		switch (mem[codeptr])
		{
			case NEWLINE_T:
			{
				codeptr++;
				if (currentpos+hugo_textwidth(pbuffer)!=0)
					AP("");
				if (mem[codeptr]==SEMICOLON_T) codeptr++;
				continue;
			}

			case TO_T:
			{
				codeptr++;

				if ((a = GetValue()) > physical_windowwidth/FIXEDCHARWIDTH)
					a = physical_windowwidth/FIXEDCHARWIDTH;

				strcpy(line, "");
				l = 0;
				if (a*FIXEDCHARWIDTH >
					hugo_textwidth(pbuffer)+currentpos-hugo_charwidth(' '))
				{
					for (i=hugo_textwidth(pbuffer)+currentpos;
				        	i<a*FIXEDCHARWIDTH;
						i+=hugo_charwidth(' '))
					{
						line[l++] = FORCED_SPACE;
						line[l] = '\0';
					}
				}
				break;
			}

			case CAPITAL_T:
			{
				codeptr++;
				capital = 1;
				continue;
			}

			case NUMBER_T:
			{
				codeptr++;
				number = 1;
				continue;
			}

			case HEX_T:
			{
				codeptr++;
				number = 1;
				hexnumber = 1;
				continue;
			}

/* PARSE_STRING_VAL and SERIAL_STRING_VAL can now be passed to GetWord()
   as constants returned by GetValue()

			case PARSE_T:
			{
				codeptr++;
				strcpy(line, parseerr);
				break;
			}

			case SERIAL_T:
			{
				codeptr++;
				strcpy(line, serial);
				break;
			}
*/

			case STRINGDATA_T:
			{
				codeptr++;
				if (game_version >= 24)
					l = PeekWord(codeptr++);
				else
					l = Peek(codeptr);
				for (i=0; i<l; i++)
					line[i] = (char)(mem[++codeptr] - CHAR_TRANSLATION);
				line[i] = '\0';
				codeptr++;
				break;
			}

			/* Anything else is treated as a value */
			default:
			{
				a = GetValue();
				if (!number)
				{
					strcpy(line, GetWord(a));
				}
				else
				{
					if (!hexnumber)
					{
						if (capital)
							itoa((unsigned int)a, line, 10);
						else
							itoa(a, line, 10);
						capital = 0;
					}
					else
						sprintf(line, "%X", a);

					number = 0;
					hexnumber = 0;
				}
				break;
			}
		}

		if (mem[codeptr]==SEMICOLON_T)
		{
			codeptr++;
			strcat(line, "\\;");
		}
		if (capital)
			{capital = 0;
			line[0] = (char)toupper((int)line[0]);}

		AP(line);
	}

	codeptr++;
}


/* RUNRESTART */

int RunRestart()
{
	char *g;
	int a;
	long i = 0;

	remaining = 0;

	g = gamefile;

	/* Use save instead of game, just in case the call fails */
	if (!(save = HUGO_FOPEN(g, "rb"))) goto RestartError;

	/* Retrieve from objtable up, since code and grammar are unchanged */
	if (fseek(save, (objtable-gameseg)*16, SEEK_SET)) goto RestartError;

	i = (objtable-gameseg)*16;
	do
	{
		if ((a = fgetc(save))==EOF and ferror(save)) goto RestartError;
		mem[i++] = (unsigned char)a;
	}
	while (i < codeend);

	if (fclose(save)) FatalError(READ_E);

	defseg = arraytable;
	for (a=0; a<MAXGLOBALS; a++)
		var[a] = PeekWord(a*2);

	i = codeptr;

	if (game_version < 22)
		{passlocal[0] = objects;
		passlocal[1] = physical_windowwidth/FIXEDCHARWIDTH;}

#if defined (DEBUGGER)
	if (active_screen!=DEBUGGER)
#endif
	{
		InitGame();

		SetStackFrame(stack_depth, RUNROUTINE_BLOCK, 0, 0);

		PassLocals(0);
#if defined (DEBUGGER)
		DebugRunRoutine((long)mainaddr*address_scale);
#else
		RunRoutine((long)mainaddr*address_scale);
#endif
		retflag = 0;

		codeptr = i;

		undoptr = 0;
		undoturn = 0;
		undoinvalid = 1;
	}

	return 1;

RestartError:
	if (fclose(save)) FatalError(READ_E);
	return 0;
}


/* RUNRESTORE */

int RunRestore()
{
	char testid[3], testserial[9];
	int lbyte, hbyte;
	int j;
	unsigned int k;
	long i;

	hugo_getfilename("to restore", savefile);

#if defined (DEBUGGER)
	if (debugger_collapsing) return 1;
#endif

	if (!(save = HUGO_FOPEN(line, "r+b"))) return 0;

	/* Check ID */
	testid[0] = (char)fgetc(save);
	testid[1] = (char)fgetc(save);
	testid[2] = '\0';
	if (ferror(save)) goto RestoreError;
	if (strcmp(testid, id))
		{AP("Incorrect save file.");
		if (fclose(save)) FatalError(READ_E);
		return 0;}

	/* Check serial number */
	if (!fgets(testserial, 9, save)) goto RestoreError;
	if (strcmp(testserial, serial))
	{
		AP("Save file created by different version.");
		if (fclose(save)) FatalError(READ_E);
		return 0;
	}

	/* Restore variables */
	for (k=0; k<MAXGLOBALS+MAXLOCALS; k++)
	{
		if ((lbyte = fgetc(save))==EOF or (hbyte = fgetc(save))==EOF)
			goto RestoreError;
		var[k] = lbyte + hbyte * 256;
	}

	/* Restore undo data */
	for (k=0; k<MAXUNDO; k++)
	{
		for (j=0; j<5; j++)
		{
			if ((lbyte = fgetc(save))==EOF or (hbyte = fgetc(save))==EOF)
				goto RestoreError;
			undostack[k][j] = lbyte + hbyte*256;
		}
	}
	if ((lbyte = fgetc(save))==EOF or (hbyte = fgetc(save))==EOF) goto RestoreError;
	undoptr = lbyte + hbyte*256;
	if ((lbyte = fgetc(save))==EOF or (hbyte = fgetc(save))==EOF) goto RestoreError;
	undoturn = lbyte + hbyte*256;
	if ((undoinvalid = (char)fgetc(save))==EOF or (undorecord = (char)fgetc(save))==EOF)
		goto RestoreError;


	/* Restore objtable and above */

	if (fseek(game, objtable*16L, SEEK_SET)) goto RestoreError;
	i = 0;

	while (i<codeend-objtable*16L)
	{
		if ((hbyte = fgetc(save))==EOF and ferror(save)) goto RestoreError;

		if (hbyte==0)
		{
			if ((lbyte = fgetc(save))==EOF and ferror(save)) goto RestoreError;
			mem[objtable*16L+i] = (unsigned char)lbyte;
			i++;

			/* Skip byte in game file */
			if (fgetc(game)==EOF) goto RestoreError;
		}
		else
		{
			while (hbyte--)
			{
				/* Get unchanged game file byte */
				if ((lbyte = fgetc(game))==EOF) goto RestoreError;
				mem[objtable*16L+i] = (unsigned char)lbyte;
				i++;
			}
		}
	}

	if (fclose(save)) FatalError(READ_E);

	strcpy(savefile, line);

	return(1);

RestoreError:
	if (fclose(save)) FatalError(READ_E);
	return 0;
}


/* RUNROUTINE

	This is the main loop for running each line of code in sequence;
	the main switch statement is based on the first token in each line.

	This routine is relatively complex, especially given the addition
	of debugger control.  Basically it is structured like this:

		1.  If this is the debugger build, see what debugger
		    information has to be set up upon calling this block
		    of code

		2.  Get the next token, and as long as it isn't
		    CLOSE_BRACE_T ('}')...

		3.  ...If this is the debugger build, see if there is a
		    standing debugger_interrupt to pass control back to
		    the debugger, and perform all operations for stepping
		    tracing, breakpoints, etc.

		4.  ...See what token we're dealing with and execute
		    accordingly

		5.  ...Loop back to (2)

		6.  If this is the debugger build, do whatever is
		    necessary to tidy up after finishing this block of
		    code

	There's a bit of a trick involved since the original language
	design uses "{...}" structures for both conditionals and blocks
	that necessitate another (i.e., nested) call to RunRoutine().
	The call_block structure array and stack_depth variable are
	the navigation guides.
*/

void RunRoutine(long addr)
{
	char null_count;        /* for reading to next address boundary     */
	char tempinexpr;
	int i, t, len, xpos, ypos;
	int initial_stack_depth, tempret;
	unsigned int routineptr = 0;
	long textaddr;
#if defined (DEBUGGER)
	char wascalled = 0;     /* distinguish routine calls from 'window'
				   blocks, etc. */
	char *called_from;
	char trace_comp_prop = 0;
	unsigned char param_type;
	long param_start;
	unsigned int old_currentroutine = 0;

	/* Because <debugdata#> and <label#> in-code data is
	   not decompiled
	*/
	int broke_on_nonstatement = 0;
#endif

	/* If routine doesn't exist */
	if (addr==0L) return;

	initial_stack_depth = stack_depth;
	inexpr = 0;

#if !defined (DEBUGGER)
#if defined (DEBUG_CODE)
/*
	if (trce)
	{
		if (codeptr != addr)
		{
			sprintf(line, "[ROUTINE:  $%6s]", PrintHex(addr));
			AP(line);
			wascalled = 1;
		}
	}
*/
#endif
#endif

#if defined (DEBUGGER)

/*
 *      First see what debugger information has to be set up upon calling
 *      this block of code
 */

	/* If this is a routine call vs. other block, codeptr will be
	   different from addr.
	*/
	if (codeptr != addr)
	{
		wascalled = 1;

		/* Checking to see if currentroutine is 0 is a way of seeing
		   if the debugger has started properly.  If, for example, a
		   property routine runs while LoadGame() is searching for
		   the display object, it may corrupt the uninitialized
		   debugger arrays.
		*/
		if ((old_currentroutine = currentroutine)==0) return;

		currentroutine = (unsigned int)(addr/address_scale);

		if (debugger_step_over)
			step_nest++;
		else
		{
			/* Add a blank line if one hasn't been added
			   already:
			*/
			if ((window[CODE_WINDOW].count) and (codeline[window[CODE_WINDOW].count-1][0]&0x0FF)!='\0')
				AddStringtoCodeWindow("");

			/* If this is a property routine, the line array
			   already holds the calling information
			*/
			if (!trace_complex_prop_routine)
				sprintf(line, "Calling:  %s", RoutineName(currentroutine));
			else
				trace_comp_prop = true;
			trace_complex_prop_routine = false;

			call[window[VIEW_CALLS].count].addr = currentroutine;
			call[window[VIEW_CALLS].count++].param = (char)arguments_passed;
			window[VIEW_CALLS].changed = true;

			/* Revise call history */
			if (window[VIEW_CALLS].count==MAXCALLS)
			{
				for (i=0; i<MAXCALLS-1; i++)
				{
					call[i].addr = call[i+1].addr;
					call[i].param = call[i+1].param;
				}
				call[0].addr = 0;
			}

			/* If not object.property or an event */
			if (strchr(line, '.')==NULL and strstr(line, "vent ")==NULL)
			{
				strcat(line, "(");
				for (i=0; i<arguments_passed; i++)
				{
					sprintf(line+strlen(line), "%d", var[MAXGLOBALS+i]);
					if (i<arguments_passed-1)
						strcat(line, ", ");
				}
				strcat(line, ")");
			}
			AddStringtoCodeWindow(line);
		}
	}

#endif  /* defined (DEBUGGER) */

	defseg = gameseg;
	codeptr = addr;

/*
 *      Get the next token, so long as it isn't a CLOSE_BRACE_T ('}')
 *      marking the end of this block of code
 */

ContinueRunning:

	while (mem[codeptr] != CLOSE_BRACE_T)   /* until "}" */
	{

#if defined (DEBUGGER)
		/* Check if we're stepping over, and if we've returned to
		   the original level of nesting:
		*/
		if (debugger_step_over and step_nest==0)
		{
			debugger_step_over = false;
			debugger_interrupt = true;
		}
#endif
		if (var[endflag]) return;

		null_count = 0;

		/* Read the next token */
		while ((t = mem[codeptr])==0)
		{
			codeptr++;

			/* Allow for padding zeroes.  If address_scale
			   zeroes are processed, we can't simply be
			   eating up the null space before an address
			   boundary.
			*/
			if (++null_count > address_scale)
				FatalError(UNKNOWN_OP_E);
		}

#if !defined (DEBUGGER)
#if defined (DEBUG_CODE)
		if (!inwindow)
		{
			sprintf(line, "[%6s:  %s]", PrintHex(codeptr), token[t]);
			AP(line);
		}
#endif
#endif

		if (game_version < 22) if (t==TEXT_T) t = TEXTDATA_T;

#if defined (DEBUGGER)
		if (++runaway_counter>=32767 and runtime_warnings)
		{
			sprintf(line, "Possible runaway loop (32767 unchecked steps)");
			RuntimeWarning(line);
			buffered_code_lines = FORCE_REDRAW;
			runaway_counter = 0;
		}

		if (t!=DEBUGDATA_T and t!=LABEL_T and !debugger_step_over)
			AddLinetoCodeWindow(codeptr);

		if ((i = IsBreakpoint(codeptr)))
		{
			if (t==DEBUGDATA_T or t==LABEL_T)
				broke_on_nonstatement = i;

			debugger_interrupt = true;

			/* '<' for "<Unknown>" */
			if (breakpoint[--i].in[0]=='<')
			{
				breakpoint[i].in = RoutineName(currentroutine);
				window[VIEW_BREAKPOINTS].changed = true;
			}
/*
				breakpoint[i].in = RoutineName((unsigned int)(codeptr/address_scale));
*/
		}

		/* Don't add in-code data to the code window */
		if (t==DEBUGDATA_T or t==LABEL_T) goto ProcessToken;


		/* Evaluate (only) any watch expressions set to break
		   when true:
		*/
		for (i=0; i<(int)window[VIEW_WATCH].count; i++)
		{
			if (watch[i].isbreak)
			{
				SetupWatchEval(i);
				if (EvalWatch())
				{
					debugger_interrupt = true;
					break;
				}
			}
		}

		/* Always update the watch window */
		window[VIEW_WATCH].changed = true;

/*
 *      Immediately following is the main Debugger interrupt point.
 *      No matter where the engine is while running, if debugger_interrupt
 *      is ever set during the execution of a command, execution is
 *      paused and control passed to the Debugger.
 */

		if (debugger_interrupt)
		{
			if (broke_on_nonstatement)
			{
				broke_on_nonstatement--;
				breakpoint[broke_on_nonstatement].addr = codeptr;
				window[VIEW_BREAKPOINTS].changed = true;
				broke_on_nonstatement = false;
			}

			if (debugger_step_over)
				AddStringtoCodeWindow("...");
			Debugger();
		}


		/* Now, additional processing for flags that may have been
		   set while execution was suspended:
		*/

		/* Collapsing the RunRoutine() call stack */
		if (debugger_collapsing) return;


		/* May be necessary to reset this if, for some
		   reason, the line array was altered (see above)
		*/
		if (!trace_complex_prop_routine)
			sprintf(line, "Calling:  %s", RoutineName(currentroutine));
		trace_complex_prop_routine = false;


		/* Add this statement to the code history */
		if (!debugger_step_back and !(debugger_step_over and step_nest>0))
		{
			if (++history_count>=MAX_CODE_HISTORY)
				 history_count = MAX_CODE_HISTORY;
			code_history[history_last] = codeptr;
			dbnest_history[history_last] = dbnest;
			if (++history_last >= MAX_CODE_HISTORY)
				history_last = 0;
		}

		/* If skipping next or stepping back */
		if (debugger_skip or debugger_step_back)
		{
			/* Debugger() has reset codeptr to next_codeptr */
			debugger_skip = false;
			continue;
		}

#endif  /* defined (DEBUGGER) */


/*
 *      This is the heart of RunRoutine():  the switch statement
 *      that executes the next engine operation based on what token
 *      has been read
 */
#if defined (DEBUGGER)
ProcessToken:
#endif
		switch (t)
		{
			/* First process any encoded, non-executable data: */

			/* If this is v2.5 or later, the compiler will have
			   noted the nesting level of this label to
			   reconcile stack_depth
			*/
			case LABEL_T:
				stack_depth = initial_stack_depth + mem[++codeptr];
				stack_depth++;
				codeptr++;
				break;

			case DEBUGDATA_T:
			{
			  switch (mem[++codeptr])
			  {
				case VAR_T:     /* local variable name */
				{
					len = mem[++codeptr];
#if defined (DEBUGGER)
					if (!debugger_has_stepped_back)
					{
						for (i=0; i<len; i++)
							localname[current_locals][i] = mem[codeptr+i+1];
						localname[current_locals][len] = '\0';
						if (++current_locals==MAXLOCALS) current_locals--;
						window[VIEW_LOCALS].count = current_locals;
					}
#endif
					codeptr+=(len+1);
					break;
				}
			  }
			  break;
			}


			/* Then the executable statements: */

			case TEXTDATA_T:        /* printed text from file */
			{
				textaddr = Peek(codeptr+1)*65536L+(long)PeekWord(codeptr+2);
				strcpy(line, GetText(textaddr));
				codeptr += 4;
				if (Peek(codeptr)==SEMICOLON_T)
					{strcat(line, "\\;");
					codeptr++;}
				if (capital)
					{line[0] = (char)toupper((int)line[0]);
					capital = 0;}
				AP(line);
				break;
			}

			case TEXT_T:
			{
				if (mem[++codeptr]==TO_T)
				{
					codeptr++;
#if defined (DEBUGGER)
					param_type = mem[codeptr];
					param_start = codeptr;
#endif
					textto = GetValue();
					if (game_version>=23) codeptr++; /* eol */
#if defined (DEBUGGER)
					/* Check if textto is 0 but was not
					   really "text to 0", but rather
					   something that evaluated to 0
					*/
					if (textto==0 and runtime_warnings)
					{
					  if (param_type!=VALUE_T or param_start!=codeptr-4)
					    RuntimeWarning("Text array address evaluates to zero");
					}
#endif
				}
				else
				{
					SetupDisplay();
				}
				break;
			}

			case MINUS_T:           /* "--" */
			case PLUS_T:            /* "++" */
				GetValue();
				codeptr++; /* eol */
				break;

			case PRINT_T:
				RunPrint();
				break;

			case PRINTCHAR_T:
			{
Printcharloop:
				codeptr++;
				i = GetValue();
				if (capital) sprintf(line, "%c\\;", toupper(i));
				else sprintf(line, "%c\\;", i);
				capital = 0;
				AP(line);
				if (Peek(codeptr)==COMMA_T)
					goto Printcharloop;
				if (game_version>=23) codeptr++; /* eol */
				break;
			}

			case STRING_T:
				RunString();
				break;

			case WINDOW_T:
				RunWindow();
				break;

			case LOCATE_T:
			{
				codeptr++;

				Flushpbuffer();

				xpos = GetValue();
				if (xpos > physical_windowwidth/FIXEDCHARWIDTH)
					xpos = physical_windowwidth/FIXEDCHARWIDTH;

				if (Peek(codeptr)==COMMA_T)
					{codeptr++;
					ypos = GetValue();}
				else
					ypos = currentline;

				if (ypos > physical_windowheight/lineheight)
					ypos = physical_windowheight/lineheight;

				hugo_settextpos(xpos, ypos);

				currentpos = (xpos-1)*FIXEDCHARWIDTH;
				currentline = ypos;

				full = ypos - 1;

				codeptr++;      /* skip EOL */
				break;
			}

			case SELECT_T:
				codeptr++;
				break;

			case CASE_T:
			case IF_T:
			case ELSEIF_T:
			case ELSE_T:
			case WHILE_T:
			case FOR_T:
				RunIf(0);
				break;

			case DO_T:
				RunDo();
				break;

			case RUN_T:
				codeptr++;
				tempret = ret;
				GetValue();     /* object.property to run */
				ret = tempret;
				if (game_version>=23) codeptr++; /* eol */
				break;

			case BREAK_T:
			{
				for (; stack_depth>0; stack_depth--)
				{
					if (code_block[stack_depth].brk)
					{
						codeptr = code_block[stack_depth].brk;
						--stack_depth;
#if defined (DEBUGGER)
						dbnest = code_block[stack_depth].dbnest;
#endif
						goto LeaveBreak;
					}
				}
				codeptr++;
LeaveBreak:
				break;
			}

			case RETURN_T:
			{
				codeptr++;
				i = inexpr;	/* don't reuse tempinexpr */
				inexpr = 1;
				SetupExpr();
				inexpr = (char)i;
				ret = EvalExpr(0);
				retflag = true;
				goto LeaveRunRoutine;
			}

			case JUMP_T:
			{
				codeptr = (long)PeekWord(codeptr + 1)*address_scale;
#if defined (DEBUGGER)
				dbnest = 0;	/* prevent "false" nesting */
#endif
				break;
			}

			case PARENT_T:
			case SIBLING_T:
			case CHILD_T:
			case YOUNGEST_T:
			case ELDEST_T:
			case YOUNGER_T:
			case ELDER_T:
			{
				inobj = true;

				/* Note:  GetValue() would actually get
				   the property/attribute to be set
				*/
				RunSet(GetVal());

				inobj = false;
				break;
			}

			case VAR_T:
			case OBJECTNUM_T:
			case VALUE_T:
			case WORD_T:
			case ARRAYDATA_T:
			case ARRAY_T:
				RunSet(-1);
				break;

			case ROUTINE_T:
			case CALL_T:
			{
				switch (t)
				{
					case ROUTINE_T:
					{
						codeptr++;
						routineptr = PeekWord(codeptr);
						codeptr += 2;
						break;
					}
					case CALL_T:
					{
						codeptr++;
						routineptr = GetValue();
					}
				}

				tempret = ret;
				CallRoutine(routineptr);
				ret = tempret;

				if (mem[codeptr]==DECIMAL_T or mem[codeptr]==IS_T)
					RunSet(ret);

				else if ((t==CALL_T) and game_version>=23)
					codeptr++; /* eol */

				break;
			}

			case MOVE_T:
			case REMOVE_T:
				RunMove();
				break;

			case COLOR_T:
			case COLOUR_T:
			{
				codeptr++;

				/* Get foreground color */
				fcolor = (char)GetValue();

				/* If background color is given */
				if (Peek(codeptr)==COMMA_T)
				{
					codeptr++;
					bgcolor = (char)GetValue();

					/* If input color is given */
					if (Peek(codeptr)==COMMA_T)
					{
						codeptr++;
						icolor = (char)GetValue();
					}
				}
				else icolor = fcolor;

				/* Only set the actual pen color now if
				   there is no text buffered
				*/
				if (pbuffer[0]=='\0')
				{
					hugo_settextcolor(fcolor);
					hugo_setbackcolor(bgcolor);
				}

				codeptr++;      /* skip EOL */
				break;
			}

			case PAUSE_T:
			{
				codeptr++;
				Flushpbuffer();
				z = hugo_waitforkey();
				wd[0] = (unsigned int)z;
				full = 0;
				break;
			}

			case RUNEVENTS_T:
				codeptr++;
				RunEvents();
				break;

			case QUIT_T:
				var[endflag] = -1;
				break;

			case INPUT_T:
				RunInput();
				codeptr++;
				break;

			case SYSTEM_T:
				RunSystem();
				if (game_version>=23) codeptr++; /* eol */
				break;

			case CLS_T:
			{
				hugo_settextcolor(fcolor);
				hugo_setbackcolor(bgcolor);
				hugo_clearwindow();
				hugo_settextpos(1, physical_windowheight/lineheight);

				if (!inwindow)
				{
					full = 0;
					default_bgcolor = bgcolor;
				}

				codeptr++;
				pbuffer[0] = '\0';
				break;
			}

			case WRITEFILE_T:
			case READFILE_T:
				FileIO();
				break;

			case WRITEVAL_T:
			{
Writevalloop:
				codeptr++;
				i = GetValue();
				if (ioblock)
				{
					if ((ioblock==2) or fprintf(io, "%c%c", i%256, i/256)<0)
						{ioerror = true;
						retflag = true;
						break;}
				}

				if (Peek(codeptr)==COMMA_T)
					goto Writevalloop;

				if (game_version>=23) codeptr++; /* eol */
				break;
			}

			case PICTURE_T:
				DisplayPicture();
				break;

			case MUSIC_T:
				PlayMusic();
				break;

			case SOUND_T:
				PlaySample();
				break;

			/* Didn't match a command token, so throw up an
			   "Unknown operation" error.
			*/
			default:
				FatalError(UNKNOWN_OP_E);
		}

		defseg = gameseg;

		if (retflag) goto LeaveRunRoutine;
	}


	/* Process the closing '}': */

	codeptr++;

#if defined (DEBUGGER)
	if (--dbnest < 0) dbnest = 0;
#endif
	/* Continue executing this iteration of RunRoutine() if the
	   '}' marks the end of a conditional block, i.e., one that
	   didn't call RunRoutine() the way, e.g., 'window' does.
	   Otherwise, get out of RunRoutine().
	*/
	if (code_block[stack_depth--].type > RUNROUTINE_BLOCK)
	{
		/* Skip a following 'elseif', 'else', or 'case'  */
		t = mem[codeptr];
		while (t==ELSEIF_T or t==ELSE_T or t==CASE_T)
		{
			RunIf(1);
			t = mem[codeptr];
		}

		if (t==WHILE_T and code_block[stack_depth+1].type==DOWHILE_BLOCK)
		{
			codeptr+=3;
			tempinexpr = inexpr;
			inexpr = 1;
			SetupExpr();
			inexpr = tempinexpr;

			if (EvalExpr(0))
				codeptr = code_block[++stack_depth].returnaddr;
			else
				codeptr = code_block[stack_depth+1].brk;
		}

		/* Since this isn't a RUNROUTINE_BLOCK, keep running this
		   iteration of RunRoutine()
		*/
		goto ContinueRunning;

	}

	if (stack_depth<0) stack_depth = 0;

	if (var[endflag]) return;


LeaveRunRoutine:

#if defined (DEBUGGER)

/*
 *      Finally, do any debugger-required cleaning-up
 */

	/* As noted above, wascalled is true if this was a routine call
	   as opposed to, e.g., a conditional block.  In the former case,
	   it is necessary to print a the "Returning from..." message.
	*/
	if (wascalled)
	{
		if (debugger_step_over)
		{
			if (--step_nest<=0)
			{
				debugger_step_over = false;
				debugger_interrupt = true;

				if (debugger_finish or step_nest < 0)
				{
					debugger_finish = false;
					goto ReturnfromRoutine;
				}
			}
		}

		else if (!debugger_step_over)
		{
ReturnfromRoutine:
			sprintf(line, "(Returning %d", ret);

			/* Since a complex property routine will give "<Routine>" as the
			   routine name, skip those
			*/
			called_from = RoutineName(currentroutine);
			if (!trace_comp_prop and called_from[0]!='<')
				sprintf(line+strlen(line), " from %s", called_from);

			if (old_currentroutine!=mainaddr and old_currentroutine!=initaddr
				and currentroutine!=mainaddr and currentroutine!=initaddr)
/*
				and currentroutine!=old_currentroutine)
*/
			{
				sprintf(line+strlen(line), " to %s", RoutineName(old_currentroutine));
			}
			strcat(line, ")");
			AddStringtoCodeWindow(line);
			AddStringtoCodeWindow("");

			if (--window[VIEW_CALLS].count < 0) window[VIEW_CALLS].count = 0;
			window[VIEW_CALLS].changed = true;
		}

		currentroutine = old_currentroutine;
	}

/*#elif defined (DEBUG_CODE)
	if (wascalled)
		{sprintf(line, "[RETURNING %d]", ret);
		AP(line);}
*/
#endif

	return;
}


/* RUNSAVE */

int RunSave()
{
	int c, j;
	int lbyte, hbyte;
	long i;
	int samecount = 0;

	hugo_getfilename("to save", savefile);

#if defined (DEBUGGER)
	if (debugger_collapsing) return 1;
#endif

	if (!strcmp(line, gamefile)) return 0;

	if (!hugo_overwrite(line)) return (0);

	if (!(save = HUGO_FOPEN(line, "w+b"))) return 0;

	/* Write ID */
	if (fputs(id, save)==EOF) goto SaveError;

	/* Write serial number */
	if (fputs(serial, save)==EOF) goto SaveError;

	/* Save variables */
	for (c=0; c<MAXGLOBALS+MAXLOCALS; c++)
	{
		hbyte = var[c] / 256;
		lbyte = var[c] - hbyte * 256;
		if (fputc(lbyte, save)==EOF or fputc(hbyte, save)==EOF) goto SaveError;
	}

	/* Save undo data */
	for (c=0; c<MAXUNDO; c++)
	{
		for (j=0; j<5; j++)
		{
			hbyte = undostack[c][j] / 256;
			lbyte = undostack[c][j] - hbyte*256;
			if (fputc(lbyte, save)==EOF or fputc(hbyte, save)==EOF)
				goto SaveError;
		}
	}
	if (fputc(undoptr-(undoptr/256)*256, save)==EOF or fputc(undoptr/256, save)==EOF)
		goto SaveError;
	if (fputc(undoturn-(undoturn/256)*256, save)==EOF or fputc(undoturn/256, save)==EOF)
		goto SaveError;
	if (fputc(undoinvalid, save)==EOF or fputc(undorecord, save)==EOF)
		goto SaveError;


	/* Save objtable to end of code space */

	if (fseek(game, objtable*16L, SEEK_SET)) goto SaveError;

	for (i=0; i<=codeend - objtable*16L; i++)
	{
		if ((lbyte = fgetc(game))==EOF) goto SaveError;
		hbyte = mem[objtable*16L+i];

		/* If memory same as original game file */
		if (lbyte==hbyte and samecount<255) samecount++;

		/* If memory differs (or samecount exceeds 1 byte) */
		else
		{
			if (samecount)
				if (fputc(samecount, save)==EOF) goto SaveError;

			if (lbyte!=hbyte)
			{
				if (fputc(0, save)==EOF) goto SaveError;
				if (fputc(hbyte, save)==EOF) goto SaveError;
				samecount = 0;
			}
			else samecount = 1;
		}
	}
	if (samecount)
		if (fputc(samecount, save)==EOF) goto SaveError;

	if (fclose(save)) FatalError(WRITE_E);

	strcpy(savefile, line);

	return(1);

SaveError:
	if (fclose(save)) FatalError(WRITE_E);
	return 0;
}


/* RUNSCRIPTSET */

int RunScriptSet()
{
	remaining = 0;

	switch (Peek(codeptr))
	{
		case SCRIPTON_T:
		{
			if (!script)
			{
				hugo_getfilename("to begin transcription (or printer name)", scriptfile);
#if defined (DEBUGGER)
				if (debugger_collapsing) return 1;
#endif
				if (!(script = HUGO_FOPEN(line, "wt")))
					return (0);
				strcpy(scriptfile, line);
				return 1;
			}
			break;
		}

		case SCRIPTOFF_T:
		{
			if (script)
			{
				if (fclose(script)) return (0);
				script = 0;
				return 1;
			}
			break;
		}
	}
	return 0;
}


/* RUNSTRING

	As in 'x = string(<array>, "<string>"[, maxlen]'.
*/

int RunString()
{
	int i;
	unsigned int aaddr;                     /* array address   */
	unsigned int dword;                     /* dictionary word */
	unsigned int maxlen = 32767;

	codeptr += 2;                           /* skip "(" */

	aaddr = GetValue();
	if (game_version>=22)
	{
		/* Convert to 16-bit word value */
		aaddr*=2;

		if (game_version>=23)
		{
			defseg = arraytable;
			maxlen = PeekWord(aaddr);
			defseg = gameseg;

			/* Space for array length */
			aaddr+=2;
		}
	}

	if (Peek(codeptr)==COMMA_T) codeptr++;

/* No longer needed--circumvented with PARSE_STRING_VAL

	if (Peek(codeptr)==PARSE_T)
		{p = 1;
		codeptr++;}
	else
*/
	dword = GetValue();

	if (Peek(codeptr)==COMMA_T) codeptr++;

	if (Peek(codeptr)!=CLOSE_BRACKET_T)
		maxlen = GetValue();
	if (Peek(codeptr)==CLOSE_BRACKET_T) codeptr++;

/*
	if (p) strcpy(line, parseerr);
	else
*/
	strcpy(line, GetWord(dword));

	defseg = arraytable;
	for (i=0; i<(int)strlen(line) and i<(int)maxlen; i++)
	{
		SaveUndo(ARRAYDATA_T, aaddr, i, PeekWord(aaddr+i*2), 0);
		PokeWord(aaddr+i*2, line[i]);
	}

	PokeWord(aaddr+i*2, 0);

	defseg = gameseg;

	return (i);
}


/* RUNSYSTEM */

#define STAT_UNAVAILABLE	((short)-1)

int RunSystem(void)
{
#if defined (GCC_OS2)
    int bkey;
#endif
	codeptr++;

	/* Since the obsolete form of the system command is unimplemented, 
	   simply get the parameter (in order to skip it), and exit the
	   function.
	*/
	if (game_version < 25)
	{
		GetValue();
		return 0;
	}
	
	/* Otherwise, process the following system calls: */
	
	codeptr++;			/* skip opening bracket */
	
	var[system_status] = 0;
	
	switch(GetValue())
	{
		case 11:		/* READ_KEY */

/* This had to be modified due to the nature of _read_kbd()
   from the EMX library, which almost but not quite duplicates
   kbhit() and getch() */

#if !defined (GCC_OS2)
			if (!hugo_iskeywaiting())
				return 0;
			else
				return hugo_getkey();
#else
            if ((bkey = _read_kbd(0, 0, 0)) == -1)
                return 0;
            else
                return hugo_getkey(bkey);
#endif

		case 21:		/* NORMALIZE_RANDOM */
			srand(1);
			break;
		case 22:		/* INIT_RANDOM */
			srand((unsigned int)time((time_t *)&z));
			break;
			
		case 31:		/* PAUSE_SECOND */
			if (!hugo_timewait(1))
				var[system_status] = STAT_UNAVAILABLE;
			break;
			
		case 32:		/* PAUSE_100TH_SECOND */
			if (!hugo_timewait(100))
				var[system_status] = STAT_UNAVAILABLE;
			break;
	}
	
	return 0;
}


/* RUNWINDOW

	As in 'window [n[, o, p, q]]'.
*/

struct SAVED_PHYSICAL_WINDOW
{
	int left, top, right, bottom;
	int width, height, charwidth, lineheight;
} saved_physical_window;

void SavePhysicalWindowData(struct SAVED_PHYSICAL_WINDOW *spw)
{
	spw->left = physical_windowleft;
	spw->top = physical_windowtop;
	spw->right = physical_windowright;
	spw->bottom = physical_windowbottom;
	spw->width = physical_windowwidth;
	spw->height = physical_windowheight;
	spw->charwidth = charwidth;
	spw->lineheight = lineheight;
}

void RestorePhysicalWindowData(struct SAVED_PHYSICAL_WINDOW *spw)
{
	physical_windowleft = spw->left;
	physical_windowtop = spw->top;
	physical_windowright = spw->right;
	physical_windowbottom = spw->bottom;
	physical_windowwidth = spw->width;
	physical_windowheight = spw->height;
	charwidth = spw->charwidth;
	lineheight = spw->lineheight;
}

void RunWindow(void)
{
	int top, bottom, left, right;
	struct SAVED_PHYSICAL_WINDOW restorewindow;
	int tempfont;
	int temp_current_text_y = current_text_y;
	char restore_default_bgcolor;
	int tempfull;
	int temp_stack_depth = stack_depth;
	FILE *tempscript;

#if defined (DEBUGGER)
	unsigned char param_type;
	int tempdbnest;
	long param_start;
#endif


	Flushpbuffer();
	tempfull = full;
	full = 0;

	tempscript = script;
	script = false;
	restore_default_bgcolor = default_bgcolor;

	/* v2.4 is the first version to support proper windowing */
	if (game_version>=24)
	{
		/* Set up default top, left, etc. as character coordinates,
		   and save the current physical window data
		*/
		left = physical_windowleft/FIXEDCHARWIDTH + 1;
		top = physical_windowtop/FIXEDLINEHEIGHT + 1;
		right = physical_windowright/FIXEDCHARWIDTH + 1;
		bottom = physical_windowbottom/FIXEDLINEHEIGHT + 1;

		SavePhysicalWindowData(&restorewindow);

		/* if "window x1, y1, x2, y2" or "window n"... */
		if (mem[++codeptr]!=EOL_T)
		{
#if defined (DEBUGGER)
			param_type = mem[codeptr];
			param_start = codeptr;
#endif
			left = GetValue();
			if (mem[codeptr++]==COMMA_T)
			{
				top = GetValue();
				if (mem[codeptr++]==COMMA_T)
				{
					right = GetValue();
					if (mem[codeptr++]==COMMA_T)
					{
						bottom = GetValue();
						codeptr++;
					}
				}
			}

			/* if only one parameter, i.e., "window n" */
			else
			{
				if (left!=0)
				{
					bottom = left;
					top = 1;
					left = 1;
					right = SCREENWIDTH;
				}

				/* "window 0" restores full screen without
				   running a code block
				*/
				else
				{
#if defined (DEBUGGER)
					/* Here, check to see if left was 0 but the
					   statement wasn't really "window 0", but
					   rather something that evaluated to zero
					*/
					if (runtime_warnings)
					{
					  if (param_type!=VALUE_T or param_start!=codeptr-4)
					    RuntimeWarning("Window size evaluates to zero");
					}
#endif
					tempfont = currentfont;
					left = 1, top = 1;
					right = SCREENWIDTH/FIXEDCHARWIDTH;
					bottom = SCREENHEIGHT/FIXEDLINEHEIGHT;
					physical_lowest_windowbottom = lowest_windowbottom = 0;
					hugo_settextwindow(left, top, right, bottom);
					goto LeaveWindow;
				}
			}
		}

		/* ...or just "window", so use last window defaults */
		else
		{
			codeptr++;      /* skip EOL */

			left = last_window_left;
			top = last_window_top;
			right = last_window_right;
			bottom = last_window_bottom;
		}

		/* Remember, these are character/text coordinates */
		if (top < 1) top = 1;
		if (left < 1) left = 1;
		if (bottom < 1) bottom = 1;
		if (right < 1) right = 1;
		if (top > SCREENHEIGHT/FIXEDLINEHEIGHT)
			top = SCREENHEIGHT/FIXEDLINEHEIGHT;
		if (left > SCREENWIDTH/FIXEDCHARWIDTH)
			left = SCREENWIDTH/FIXEDCHARWIDTH;
		if (bottom > SCREENHEIGHT/FIXEDLINEHEIGHT)
			bottom = SCREENHEIGHT/FIXEDLINEHEIGHT;
		if (right > SCREENWIDTH/FIXEDCHARWIDTH)
			right = SCREENWIDTH/FIXEDCHARWIDTH;

		/* Set the new text window */
		hugo_settextwindow(left, top, right, bottom);
		hugo_settextpos(1, 1);

		inwindow = true;

#if defined (DEBUGGER)
		tempdbnest = dbnest++;
#endif
		SetStackFrame(stack_depth, RUNROUTINE_BLOCK, 0, 0);

		RunRoutine(codeptr);

#if defined (DEBUGGER)
		dbnest = tempdbnest;
#endif
		stack_depth = temp_stack_depth;

		Flushpbuffer();
		inwindow = false;

		/* Restore the old window parameters */
		last_window_top = top;   last_window_bottom = bottom;
		last_window_left = left; last_window_right = right;

		/* Figure out what the lowest window bottom is that we need
		   to protect from scrolling
		*/
		if (bottom > lowest_windowbottom)
			lowest_windowbottom = bottom;

		/* (error situation--shouldn't happen) */
		if (lowest_windowbottom>=SCREENHEIGHT/FIXEDLINEHEIGHT)
			lowest_windowbottom = 0;

		/* Restore the old text window */
		tempfont = currentfont;
		RestorePhysicalWindowData(&restorewindow);

		hugo_settextwindow(physical_windowleft/FIXEDCHARWIDTH + 1,
			lowest_windowbottom + 1,
			physical_windowright/FIXEDCHARWIDTH + 1,
			physical_windowbottom/FIXEDLINEHEIGHT + 1);

		physical_lowest_windowbottom = lowest_windowbottom*FIXEDLINEHEIGHT;
	}

	/* v2.3 and earlier supported a very simple version of
	   windowing:  mainly just moving the top/scroll-off line
	   of the printable area to the bottom of the text printed
	   in the "window" block
	*/
	else
	{
		hugo_settextwindow(1, 1,
	       		SCREENWIDTH/FIXEDCHARWIDTH,
		        SCREENHEIGHT/FIXEDLINEHEIGHT);
		hugo_settextpos(1, 1);

		inwindow = true;

		SetStackFrame(stack_depth, RUNROUTINE_BLOCK, 0, 0);
		RunRoutine(++codeptr);
		Flushpbuffer();

		inwindow = false;

		stack_depth = temp_stack_depth;

		hugo_settextwindow(1, full+1,
		        SCREENWIDTH/FIXEDCHARWIDTH,
			SCREENHEIGHT/FIXEDLINEHEIGHT);

		physical_lowest_windowbottom = full*lineheight;
	}

LeaveWindow:

	hugo_settextpos(1, physical_windowheight/lineheight);
	current_text_y = temp_current_text_y;

	default_bgcolor = restore_default_bgcolor;
	script = tempscript;

	full = tempfull;
}

